#include <stdint.h>
#include <xen/xen.h>
#include "console.h"
#include "hypercall.h"
#include "mem.h"

/*finds. the length of the strings
*Input:A pointer to string
*Returns the calculated length */

int findlength(const void *str) {
  char *tmp=(char *)str;
  int i,len=0;
  while(*tmp!='\0') {
    tmp++;
    len++;
  }
  return len;
}

/*Copies the values of num bytes from the location pointed
by source directly to the memory block pointed by destination*/

void *memcpy(void *dest, const void *src, size_t n) {
  char *tmp1=(char *)dest;
  char *tmp2=(char *)src;
  int src_len=findlength(src);
  int dest_len=findlength(dest);
  if(src>dest) {
    if(dest+src_len > src) {

      console_write("Address overlap. Cannot copy.\r\n");
      return 1;
    }
  }
  else {
    if(src+src_len > dest) {
      console_write("Address overlap. Cannot copy.\r\n");
      return 1;
    }
  }
  if (n==0) return 1;
  if (n<0) {
    console_write("Length has to be positive\r\nSegmentation fault\r\n");
    return 1;
  }
  while(*tmp2!='\0' && n>0) {
    (*tmp1)=(*tmp2);
    tmp1++;
    tmp2++;
    n--;
  }
  *tmp1='\0';
}

/*Sets the first num bytes of the block of memory
pointed by ptr to the specified value*/

void *memset(void *dest, int val, size_t n) {
  char *tmp1;
  if(n==0) return 1;
  if(n<0) {
    console_write("Size argument cannot be negative\r\nSegmentation fault\r\n");
    return 1;
  }
  tmp1=(char *)dest;
  int len=findlength(tmp1);
  while (n>0) {
    (*tmp1)=val;
    tmp1++;
    n--;
    len--;
  }
  while(len>0) {
    len--;
    tmp1++;
  }
  *tmp1='\0';
}

/*Compares the first n bytes of the block of memory
pointed by s1 to the first n bytes pointed by s2,
returning zero if they all match or a value different
from zero representing which is greater if they do not*/

int memcmp(const void *s1, const void *s2, size_t n) {
  int val=0;
  char *tmp1=(char *)s1;
  char *tmp2=(char *)s2;
  while(*tmp1!='\0' || *tmp2!='\0') {
    if(n>0) {
      if (*tmp1==*tmp2) {
        tmp1++;
        tmp2++;
        n--;
      }
      else {
        val= *tmp1 - *tmp2;
        return val;
      }
    }
    else {
      break;
    }
  }
  return val;
}

/*Appends a copy of the s1 string to the s2 string.*/
char* strcat(char *s2, const char *s1) {
  char *tmp;
  int src_len=findlength(s1);
  int dest_len=findlength(s2);
  if(s1>s2) {
    if(s2+src_len > s1) {
      console_write("Address overlap. Cannot copy.\r\n");
      return 1;
    }
  }
  else {
    if(s1+src_len > s2) {
      console_write("Address overlap. Cannot copy.\r\n");
      return 1;
    }
  }
  tmp=s2;
  while(*tmp!='\0') tmp++;
  while(*s1!='\0') {
    (*tmp)=(*s1);
    tmp++;
    s1++;
  }
  *tmp='\0';
  return s2;
}

/*Appends the first n characters of s1 to s2*/
char* strncat(char *s2, const char *s1, size_t n) {
  char *tmp;
  int src_len=findlength(s1);
  int dest_len=findlength(s2);
  unsigned long num;
  /*needed because the third parameter comes in as an unsigned parameter
   if the change is not there when we input negative nos. we get incorrect results*/
  if(n<0)
    num=n;
  else
    num=n;
  if(s1>s2) {
    if(s2+src_len > s1) {
      console_write("Address overlap. Cannot copy.\r\n");
      return 1;
    }
  }
  else {
    if(s1+src_len > s2) {
      console_write("Address overlap. Cannot copy.\r\n");
      return 1;
    }
  }
  tmp=s2;
  while(*tmp!='\0')
    tmp++;
  while(*s1!='\0') {
    if(num==0)
      break;
    if(num>0) {
      num--;
      (*tmp)=(*s1);
      tmp++;
      s1++;
    }
  }
  *tmp='\0';
  return s2;
}

/*Compares the C string s1 to the C string s2.
This function starts comparing the first character
of each string. If they are equal to each other, it
continues with the following pairs until the characters
differ or until a terminating null-character is reached.*/

int strcmp(const char *s1, const char* s2) {
  int val=0;
  char *tmp1=(char *)s1;
  char *tmp2=(char *)s2;
  while(*tmp1!='\0' || *tmp2!='\0') {
    if(*tmp1==*tmp2) {
      tmp1++;
      tmp2++;
    }
    else {
      val= *tmp1 - *tmp2;
      return val;
    }
  }
  return val;
}

/*Compares up to n characters of the C string
s1 to those of the C string s2.*/

int strncmp(const char *s1, const char* s2, size_t n) {
  int val=0;
  char *tmp1=(char *)s1;
  char *tmp2=(char *)s2;
  unsigned long num;
  /*needed because the third parameter comes in as an unsigned parameter
   if the change is not there when we input negative nos. we get incorrect results*/
  if(n<0)
    num=n;
  else
    num=n;
  while(*tmp1!='\0' || *tmp2!='\0') {
    if(num>0) {
      if(*tmp1==*tmp2) {
        tmp1++;
        tmp2++;
        num--;
      }
      else {
        val= *tmp1 - *tmp2;
        return val;
      }
    }
    else {
      break;
    }
  }
  return val;
}

/*Copies the C string pointed by s1 into
the array pointed by s2, including
the terminating null character.*/

char* strcpy(char *s2, const char *s1) {
  char *tmp;
  int src_len=findlength(s1);
  int dest_len=findlength(s2);
  int len=src_len-1;
  if(s1>s2) {
    if(s2+src_len > s1) {
      console_write("Address overlap. Cannot copyhere.\r\n");
      return 1;
    }
  }
  else {
    if(s1+src_len>s2) {
      while(len>=0) {
        *(s2+len)=*(s1+len);
        len--;
      }
      *(s2+src_len)='\0';
      return s2;
    }
  }
  tmp=s2;
  while(*s1!='\0') {
    (*tmp)=(*s1);
    tmp++;
    s1++;
  }
  *tmp='\0';
  return s2;
}

/*Copies the first num characters of s1 to s2*/

char* strncpy(char *s2, const char *s1, size_t n) {
  char *tmp;
  int src_len=findlength(s1);
  int dest_len=findlength(s2);
  if(n<0) {
    console_write("Segmentation Fault.\r\n");
    console_write("Negative arguments are not allowed.\r\n");
    return 1;
  }
  if(s1>s2) {
    if(s2+src_len > s1) {
      console_write("Address overlap. Cannot copy.\r\n");
      return 1;
    }
  }
  else {
    if(s1+src_len > s2) {
      console_write("Address overlap. Cannot copy.\r\n");
      return 1;
    }
  }
  tmp=s2;
  while(*s1!='\0') {
    if(n==0)
      break;
    if (n>0) {
      dest_len--;
      n--;
    }
    (*tmp)=(*s1);
    tmp++;
    s1++;
  }
  while(dest_len>0) {
    tmp++;
    dest_len--;
  }
  *tmp='\0';
  return s2;
}

char* dec2str(int val,int base,int who) {
  char temp;
  int rem,i=0,j,neg;
  unsigned int no;
  if(val<0 && who==1) {
    //consoleout("-",1);
    console_write("-");
    no=val*-1;
  }
  else if(val<0 && who==2) {
    no=val; 
  }
  else if(val>0) {
    no=val;
  }
  while (no>0) {
    rem=no%base;
    res[i++]=rem + '0';
    no=no/10;
  }
  res[i]='\0';
  j=i-1;
  i=0;
  while(i<=j) {
    temp=res[i];
    res[i]=res[j];
    res[j]=temp;
    i++;
    j--;
  }
  return res;
}

char* ldec2str(long val,int base,int who) {
  char temp;
  int rem,i=0,j,neg;
  unsigned long no;
  if(val<0 && who==1) {
    //consoleout("-",1);
    console_write("-");
    no=val*-1;
  }
  else if(val<0 && who==2) {
    no=val; 
  }
  else if(val>0) {
    no=val;
  }
  while (no>0) {
    rem=no%base;
    res[i++]=rem + '0';
    no=no/10;
  }
  res[i]='\0';
  j=i-1;
  i=0;
  while(i<=j) {
    temp=res[i];
    res[i]=res[j];
    res[j]=temp;
    i++;
    j--;
  }
  return res;
}

char* hex2str(unsigned int val, int base) {
  char temp;
  int rem,i=0,j;
  while (val>0) {
    rem=val%base;
    if(rem <=9)
      res[i++]=rem + '0';
    else
      res[i++]=rem - 10 + 'a';
    val=val/16;
  }
  res[i]='\0';
  j=i-1;
  i=0;
  while(i<=j) {
    temp=res[i];
    res[i]=res[j];
    res[j]=temp;
    i++;
    j--;
  }
  return res;
}

char* lhex2str(unsigned long val, int base) {
  char temp;
  int rem,i=0,j;
  while (val>0) {
    rem=val%base;
    if(rem <=9)
      res[i++]=rem + '0';
    else
      res[i++]=rem - 10 + 'a';
    val=val/16;
  }
  res[i]='\0';
  j=i-1;
  i=0;
  while(i<=j) {
    temp=res[i];
    res[i]=res[j];
    res[j]=temp;
    i++;
    j--;
  }
  return res;
}

int printf(char *fmt, ...) {
  va_list ap;
  int i;
  char *str,*str1;
  char string[10];
  int val;
  int len;
  long lval;
  unsigned int hval;
  unsigned long hlval;
  char ch;
  va_start(ap,fmt);
  str1=string;
  for (; *fmt != 0; fmt++)
  {
    if(*fmt=='%') {
      fmt++;
      switch(*fmt) {
      case 's':
               str1=va_arg(ap,char*);
               console_write(str1);
               continue;
      case 'd':
               if(*(fmt+1) == 'l') {
                 fmt++; 
                 lval=va_arg(ap,long);
                 str=ldec2str(lval,10,1);
                 console_write(str);
               }
               else {
                 val=va_arg(ap,int);
                 str=dec2str(val,10,1);
                 console_write(str);
               }
               continue;
      case 'x':
               if(*(fmt+1) == 'l') {
                 fmt++;
                 hlval=va_arg(ap,unsigned long);
                 str=lhex2str(hlval,16);
                 console_write(str);
               }
               else {
                 hval=va_arg(ap,unsigned int);
                 str=hex2str(hval,16);
                 console_write(str); 
               }
               continue;
      case 'u':
               if(*(fmt+1) == 'l') {
                 fmt++;
                 lval=va_arg(ap,long);
                 str=ldec2str(lval,10,2);
                 console_write(str);
               }
               else {
                 val=va_arg(ap,int);
                 str=dec2str(val,10,2);
                 console_write(str);
               }
               continue;
      }
    }
    else {
      //consoleout(fmt,1);
      ch=*fmt;
      console_write(&ch);
      //call compile function
    }
  }
  console_write("\r\n");
  va_end(ap);
}

